// object_factory.hpp
//
//

// http://www.gamedev.net/reference/articles/article2097.asp
// http://www.boost.org/doc/libs/1_35_0/libs/preprocessor/doc/index.html

// this works by 
// - some preprocessor magic
// - explicit template argument specification for member functions

#ifndef OBJECT_FACTORY_HPP
#define OBJECT_FACTORY_HPP

#include <map>

#include <boost/preprocessor/iteration.hpp>
#include <boost/preprocessor/enum_params.hpp>
#include <boost/preprocessor/repetition/enum_binary_params.hpp>

#define LIMIT_LOWER_BOUND 1
#define LIMIT_UPPER_BOUND 10
#define BOOST_PP_LOCAL_LIMITS (LIMIT_LOWER_BOUND, LIMIT_UPPER_BOUND)

// primary template
template <typename CtorSignature, typename UniqueIdType> class ObjectFactory;

// constructor with no arguments
template< typename Base, typename Derived >
Base CreateObject()
{
  return new Derived();
}

template<typename Base, typename UniqueId >
class ObjectFactory
{
  typedef Base (*CreateObjectFunc)();
  typedef std::map<UniqueId, CreateObjectFunc> object_creator_t;
  object_creator_t object_creator_;
public:
  template<typename Derived>
  bool Register(UniqueId unique_id)
  {
    if (object_creator_.find(unique_id) != object_creator_.end())
      return false;

    object_creator_[unique_id] = 
      &CreateObject< Base, Derived >;

    return true;
  }

  Base Create(UniqueId unique_id)
  {
    typedef typename object_creator_t::const_iterator const_iterator;
    const_iterator iter = object_creator_.find(unique_id);

    if (iter == object_creator_.end())
       return 0;

    return iter->second;
  }
};

// n arguments
#define BOOST_PP_LOCAL_MACRO(n) \
template< typename Base, BOOST_PP_ENUM_PARAMS(n, typename T), typename Derived > \
Base CreateObject(BOOST_PP_ENUM_BINARY_PARAMS(n, T, t)) \
{ \
  return new Derived(BOOST_PP_ENUM_PARAMS(n, t)); \
} \
\
template<typename Base, BOOST_PP_ENUM_PARAMS(n, typename T), typename UniqueId > \
class ObjectFactory< Base( BOOST_PP_ENUM_PARAMS(n, T)), UniqueId > \
\
{ \
  typedef Base (*CreateObjectFunc)(BOOST_PP_ENUM_PARAMS(n, T)); \
  typedef std::map<UniqueId, CreateObjectFunc> object_creator_t; \
  object_creator_t object_creator_; \
public: \
  template<typename Derived> \
  bool Register(UniqueId unique_id) \
  { \
    if (object_creator_.find(unique_id) != object_creator_.end()) \
      return false; \
\
    object_creator_[unique_id] = \
      &CreateObject< Base, BOOST_PP_ENUM_PARAMS(n, T), Derived >; \
\
    return true; \
  } \
\
  Base Create(UniqueId unique_id, BOOST_PP_ENUM_BINARY_PARAMS(n, T, t)) \
  { \
    typedef typename object_creator_t::const_iterator const_iterator; \
    const_iterator iter = object_creator_.find(unique_id); \
\
    if (iter == object_creator_.end()) \
       return 0; \
\
    return iter->second(BOOST_PP_ENUM_PARAMS(n, t)); \
  } \
};

#include BOOST_PP_LOCAL_ITERATE()

#undef LIMIT_LOWER_BOUND
#undef LIMIT_UPPER_BOUND

#endif
